from abc import ABC, abstractmethod
from typing import Any


class Checker(ABC):
    """An abstract class defining the interface for a checker to be used by a Waiter."""

    @abstractmethod
    def __str__(self) -> str:
        """Textual representation of the Checker object for logging"""

    @abstractmethod
    def check(self, data: Any) -> bool:
        """Check if the given data meets a certain condition.

        Args:
            data (Any): The data to be checked.

        Returns:
            bool: True if the condition is met, False otherwise."""



class WaiterIsNotReady(Exception):
    def __init__(self, message: str = "Not all required attributes have been set!"):
        super().__init__(message)


class WaiterConditionWasNotMet(Exception):
    pass



"""A library facilitating work with asynchronous APIs."""
import logging
from logging import NullHandler

from .api import (
    RequestsWaiter,
    to_curl,
    wait_for_value_in_request,
    wait_for_values_in_request,
)
from .waiter_src.checkers import CHECKERS
from .waiter_src.comparators import COMPARATORS

__version__ = "0.3.1"
__all__ = [
    "to_curl",
    "wait_for_values_in_request",
    "wait_for_value_in_request",
    "RequestsWaiter",
    "CHECKERS",
    "COMPARATORS",
]

logging.getLogger(__name__).addHandler(NullHandler())



import logging
import uuid
from abc import abstractmethod
from json import JSONDecodeError
from typing import Any, Callable

from dictor import dictor
from requests import Response

from bepatient.waiter_src.checker import Checker

log = logging.getLogger(__name__)


class ResponseChecker(Checker):
    """A checker that compares a value in a response against an expected value.

    Args:
        comparer (Callable): A function that performs the comparison.
        expected_value (Any): The expected value to compare against."""

    def __init__(self, comparer: Callable[[Any, Any], bool], expected_value: Any):
        self.comparer = comparer
        self.expected_value = expected_value
        self._prepared_data: dict[str, Any] | list[Any] | int | None = None

    def __str__(self) -> str:
        attrs = self.__dict__.copy()
        del attrs["_prepared_data"]
        attrs["checker"] = self.__class__.__name__
        attrs["comparer"] = self.comparer.__name__

        return (
            " | ".join([f"{k.capitalize()}: {v}" for k, v in sorted(attrs.items())])
            + f" | Data: {self._prepared_data}"
        )

    @abstractmethod
    def prepare_data(self, data: Response, run_uuid: str | None = None) -> Any:
        """Prepare the data from the response for comparison.

        Args:
            data (Response): response containing the data.
            run_uuid (str | None): unique run identifier. Defaults to None.

        Returns:
            Any: Data for comparison."""

    def check(self, data: Response) -> bool:
        """Checks if the value in the JSON response satisfies the condition.

        Args:
            data (Response): response containing the JSON data.

        Returns:
            bool: True if the value in the JSON response satisfies the condition,
                False otherwise."""
        run_uuid = str(uuid.uuid4())
        log.info("Check uuid: %s | %s", run_uuid, self)

        try:
            self._prepared_data = self.prepare_data(data, run_uuid)
            if self.comparer(self._prepared_data, self.expected_value):
                return True
            log.debug(
                "Check uuid: %s | Condition not met | Expected: %s | Data: %s",
                run_uuid,
                self.expected_value,
                self._prepared_data,
            )
        except (TypeError, JSONDecodeError):
            log.exception(
                "Check uuid: %s | Expected: %s | Headers: %s | Content %s",
                run_uuid,
                self.expected_value,
                data.headers,
                data.content,
            )

        return False


class StatusCodeChecker(ResponseChecker):
    def prepare_data(self, data: Response, run_uuid: str | None = None) -> int:
        """Prepare the response status code for comparison.

        Args:
            data (Response): response containing the status code.
            run_uuid (str | None, optional): unique run identifier. Defaults to None.

        Returns:
            int: prepared status code for comparison."""
        status_code = data.status_code
        log.info("Check uuid: %s | Response status code: %s", run_uuid, status_code)
        return status_code


class DictResponseChecker(ResponseChecker):
    """A checker that compares a value in a response against an expected value.

    Args:
        comparer (Callable): A function that performs the comparison.
        expected_value (Any): The expected value to compare against.
        dict_path (str, optional): dot-separated path to the value in the response data.
            Defaults to None.
        search_query (str, optional): A search query to find the value in the response
            data. Defaults to None.
        dictor_fallback (str, optional): A default value to return if the value at the
            specified path is not found using `dictor`. Defaults to None."""

    def __init__(
        self,
        comparer: Callable[[Any, Any], bool],
        expected_value: Any,
        dict_path: str | None = None,
        search_query: str | None = None,
        dictor_fallback: str | None = None,
    ):
        self.path = dict_path
        self.search_query = search_query
        self.dictor_fallback = dictor_fallback
        super().__init__(comparer, expected_value)

    @staticmethod
    @abstractmethod
    def parse_response(
        data: Response, run_uuid: str | None = None
    ) -> dict[str, Any] | list[Any]:
        """Parse the response data into a dictionary or a list for comparison.

        Args:
            data (Response): response containing the data.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            dict[str, Any] | list[Any]: The parsed response data for comparison."""

    def prepare_data(
        self, data: Response, run_uuid: str | None = None
    ) -> dict[str, Any] | list[Any] | None:
        """Prepare the response data for comparison.

        Args:
            data (Response): The response containing the data.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            dict[str, Any] | list[Any] | None: The prepared data for comparison."""
        try:
            return dictor(
                self.parse_response(data, run_uuid),
                self.path,
                search=self.search_query,
                default=self.dictor_fallback,
            )
        except (TypeError, JSONDecodeError):
            log.exception(
                "Check uuid: %s | Expected: %s | Headers: %s | Content %s",
                run_uuid,
                self.expected_value,
                data.headers,
                data.content,
            )
        return None


class JsonChecker(DictResponseChecker):
    """A checker that compares a value in a JSON response against an expected value.

    Example:
        To check if a specific field "status" in a JSON response equals 200:

        ```
            checker = JsonChecker(lambda a, b: a == b, 200, dict_path="status")
            assert checker.check(response) is True
        ```"""

    @staticmethod
    def parse_response(
        data: Response, run_uuid: str | None = None
    ) -> dict[str, Any] | list[Any]:
        """Parse the response content as JSON for comparison.

        Args:
            data (Response): The response containing the JSON data.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            dict[str, Any] | list[Any]: The parsed JSON response data for comparison."""
        log.info("Check uuid: %s | Response content: %s", run_uuid, data.content)
        return data.json()


class HeadersChecker(DictResponseChecker):
    """A checker that compares response headers against expected values.

    Example:
        To check if the "Content-Type" header in a response is "application/json":
        ```
            checker = HeadersChecker(
                lambda a, b: a == b, "application/json", dict_path="Content-Type"
            )
            assert checker.check(response) is True
        ```
    """

    @staticmethod
    def parse_response(data: Response, run_uuid: str | None = None) -> dict[str, str]:
        """Parse the response headers for comparison.

        Args:
            data (Response): The response containing the headers.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            dict[str, str]: The parsed response headers for comparison."""
        headers = dict(data.headers)
        log.info("Check uuid: %s | Response headers: %s", run_uuid, headers)
        return headers



import pytest
from requests import get

from bepatient import (
    RequestsWaiter,
    wait_for_value_in_request,
    wait_for_values_in_request,
)


@pytest.mark.e2e
class TestRequestsWaiter:
    def test_pokeapi(self):
        msg = (
            "Moves cannot score critical hits against this Pokémon.\n\n"
            "This ability functions identically to shell armor."
        )

        waiter = RequestsWaiter(
            request=get("https://pokeapi.co/api/v2/ability/battle-armor", timeout=5)
        )
        waiter.add_checker(
            comparer="contain_all",
            expected_value=(msg,),
            dict_path="effect_entries",
            search_query="effect",
        )
        waiter.run()

        response = waiter.get_result()

        assert response.status_code == 200
        assert response.json()["name"] == "battle-armor"

    def test_pokeapi_multiple_checkers(self):
        msg = (
            "Moves cannot score critical hits against this Pokémon.\n\n"
            "This ability functions identically to shell armor."
        )
        response = (
            RequestsWaiter(
                request=get(
                    "https://pokeapi.co/api/v2/ability/battle-armor", timeout=5
                ),
            )
            .add_checker(
                expected_value=(msg,),
                comparer="contain_all",
                dict_path="effect_entries",
                search_query="effect",
            )
            .add_checker(
                checker="headers_checker",
                expected_value="cloudflare",
                comparer="is_equal",
                dict_path="Server",
            )
            .run()
            .get_result()
        )

        assert response.status_code == 200
        assert response.json()["name"] == "battle-armor"


@pytest.mark.e2e
class TestWaitForValueInRequest:
    def test_pokeapi(self):
        msg = (
            "Moves cannot score critical hits against this Pokémon.\n\n"
            "This ability functions identically to shell armor."
        )
        response = wait_for_value_in_request(
            request=get("https://pokeapi.co/api/v2/ability/battle-armor", timeout=5),
            comparer="contain_all",
            expected_value=(msg,),
            checker="json_checker",
            dict_path="effect_entries",
            search_query="effect",
        )
        assert response.status_code == 200
        assert response.json()["name"] == "battle-armor"


@pytest.mark.e2e
class TestWaitForValuesInRequest:
    def test_pokeapi(self):
        msg = (
            "Moves cannot score critical hits against this Pokémon.\n\n"
            "This ability functions identically to shell armor."
        )
        list_of_checkers = [
            {
                "checker": "json_checker",
                "comparer": "contain_all",
                "expected_value": (msg,),
                "dict_path": "effect_entries",
                "search_query": "effect",
            },
            {
                "checker": "headers_checker",
                "comparer": "is_equal",
                "expected_value": "cloudflare",
                "dict_path": "Server",
            },
        ]
        response = wait_for_values_in_request(
            request=get("https://pokeapi.co/api/v2/ability/battle-armor", timeout=5),
            checkers=list_of_checkers,  # type: ignore
            retries=5,
        )
        assert response.status_code == 200
        assert response.json()["name"] == "battle-armor"



import pytest
from requests import PreparedRequest, Response, Session

from bepatient import (
    RequestsWaiter,
    to_curl,
    wait_for_value_in_request,
    wait_for_values_in_request,
)
from bepatient.waiter_src.checkers.response_checkers import HeadersChecker
from bepatient.waiter_src.comparators import is_equal
from bepatient.waiter_src.exceptions.waiter_exceptions import WaiterConditionWasNotMet


def test_to_curl(prepared_request: PreparedRequest):
    expected_curl = (
        "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' "
        "https://webludus.pl/"
    )

    assert to_curl(prepared_request) == expected_curl


class TestRequestsWaiter:
    def test_init(self, prepared_request: PreparedRequest, session_mock: Session):
        executor = RequestsWaiter(
            request=prepared_request, status_code=201, session=session_mock
        ).executor

        assert executor.request == prepared_request
        assert executor.session == session_mock

        checker = executor._status_code_checker  # pylint: disable=protected-access
        assert checker.expected_value == 201

    def test_add_checker(self, prepared_request: PreparedRequest):
        waiter = RequestsWaiter(request=prepared_request)
        executor = waiter.add_checker(
            expected_value="TEST",
            comparer="is_equal",
            checker="headers_checker",
            dict_path="dict",
            search_query="search",
        ).executor
        w_checker = executor._checkers[0].__dict__  # pylint: disable=protected-access

        checker = HeadersChecker(
            comparer=is_equal,
            expected_value="TEST",
            dict_path="dict",
            search_query="search",
        ).__dict__

        assert w_checker == checker

    def test_happy_path(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        dict_content_response: Response,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=True, comparer="is_equal", dict_path="ok")
        response = waiter.run(retries=1).get_result()

        assert response == dict_content_response

    def test_condition_not_met_raise_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        dict_content_response: Response,
        error_msg: str,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=False, comparer="is_equal", dict_path="ok")
        with pytest.raises(WaiterConditionWasNotMet, match=error_msg):
            waiter.run(retries=1)

        assert waiter.get_result() == dict_content_response

    def test_condition_not_met_without_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        dict_content_response: Response,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=False, comparer="is_equal", dict_path="ok")
        waiter.run(retries=1, raise_error=False)

        assert waiter.get_result() == dict_content_response


class TestWaitForValueInRequests:
    def test_happy_path(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        dict_content_response: Response,
    ):
        result = wait_for_value_in_request(
            request=prepared_request,
            session=session_mock,
            comparer="is_equal",
            expected_value=True,
            dict_path="ok",
            retries=1,
        )

        assert result == dict_content_response

    def test_condition_not_met_raise_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        error_msg: str,
    ):
        with pytest.raises(WaiterConditionWasNotMet, match=error_msg):
            wait_for_value_in_request(
                request=prepared_request,
                session=session_mock,
                comparer="is_equal",
                expected_value=False,
                dict_path="ok",
                retries=1,
            )


class TestWaitForValuesInRequests:
    def test_happy_path(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        dict_content_response: Response,
    ):
        list_of_checkers = [
            {
                "checker": "json_checker",
                "comparer": "contain",
                "expected_value": "John",
                "dict_path": "list_of_dicts",
                "search_query": "name",
            },
            {
                "checker": "headers_checker",
                "comparer": "is_equal",
                "expected_value": "json",
                "dict_path": "content",
            },
        ]

        response = wait_for_values_in_request(
            request=prepared_request,
            session=session_mock,
            checkers=list_of_checkers,
            retries=1,
        )

        assert response == dict_content_response

    def test_condition_not_met_raise_error(
        self, prepared_request: PreparedRequest, session_mock: Session
    ):
        msg = (
            "The condition has not been met! | Failed checkers:"
            " (Checker: JsonChecker | Comparer: contain | Dictor_fallback: None"
            " | Expected_value: Jerry | Path: list_of_dicts | Search_query: name"
            " | Data: ['John', 'Mike'], Checker: HeadersChecker | Comparer: is_equal"
            " | Dictor_fallback: None | Expected_value: xml | Path: content"
            " | Search_query: None | Data: json) | curl -X GET -H 'task: test' -H"
            " 'Cookie: user-token=abc-123' https://webludus.pl/"
        )
        list_of_checkers = [
            {
                "checker": "json_checker",
                "comparer": "contain",
                "expected_value": "Jerry",
                "dict_path": "list_of_dicts",
                "search_query": "name",
            },
            {
                "checker": "headers_checker",
                "comparer": "is_equal",
                "expected_value": "xml",
                "dict_path": "content",
            },
        ]

        with pytest.raises(WaiterConditionWasNotMet, match=msg):
            wait_for_values_in_request(
                request=prepared_request,
                session=session_mock,
                checkers=list_of_checkers,
                retries=1,
            )



from typing import Any

from requests import PreparedRequest, Response, Session

from .curler import Curler
from .waiter_src.checkers import CHECKERS, RESPONSE_CHECKERS
from .waiter_src.comparators import COMP_DICT, COMPARATORS
from .waiter_src.executors.requests_executor import RequestsExecutor
from .waiter_src.waiter import wait_for_executor


def to_curl(req_or_res: PreparedRequest | Response, charset: str | None = None) -> str:
    """Converts a `PreparedRequest` or a `Response` object to a `curl` command.

    Args:
        req_or_res (PreparedRequest | Response): The `PreparedRequest` or `Response`
            object to be converted.
        charset (str, optional): The character set to use for encoding the
            request body, if it is a byte string. Defaults to "utf-8".

    Returns:
        the `curl` command as a string"""
    return Curler().to_curl(req_or_res, charset)


class RequestsWaiter:
    """Utility class for setting up and monitoring requests for expected values.

    Args:
        request (PreparedRequest | Response): request or response to monitor.
        status_code (int, optional): The expected HTTP status code. Defaults to 200.
        session (Session | None, optional): The requests session to use for sending
            requests. Defaults to None.

    Example:
        To wait for a JSON response where the "status" field equals 200 using a
            RequestsWaiter:
        ```
            waiter = RequestsWaiter(request=requests, status_code=200, session=session)
            response = waiter.add_checker(
                expected_value=0,
                comparer="have_len_greater",
                checker="json_checker",
                dict_path="data"
            ).run(retries=5, delay=2).get_result()
        ```"""

    def __init__(
        self,
        request: PreparedRequest | Response,
        status_code: int = 200,
        session: Session | None = None,
    ):
        self.executor = RequestsExecutor(
            req_or_res=request, expected_status_code=status_code, session=session
        )

    def add_checker(
        self,
        expected_value: Any,
        comparer: COMPARATORS,
        checker: CHECKERS = "json_checker",
        dict_path: str | None = None,
        search_query: str | None = None,
    ):
        """Add a response checker to the waiter.

        Args:
            expected_value (Any): The value to be compared against the response data.
            comparer (COMPARATORS): The comparer function or operator used for
                value comparison.
            checker (CHECKERS, optional): The type of checker to use. Defaults to
                "json_checker".
            dict_path (str | None, optional): The dot-separated path to the value in the
                response data. Defaults to None.
            search_query (str | None, optional): A search query to use to find the value
                in the response data. Defaults to None.

        Returns:
            self: updated RequestsWaiter instance."""
        self.executor.add_checker(
            RESPONSE_CHECKERS[checker](  # type: ignore
                comparer=COMP_DICT[comparer],
                expected_value=expected_value,
                dict_path=dict_path,
                search_query=search_query,
            )
        )
        return self

    def run(self, retries: int = 60, delay: int = 1, raise_error: bool = True):
        """Run the waiter and monitor the specified request or response.

        Args:
            retries (int, optional): The number of retries to perform. Defaults to 60.
            delay (int, optional): The delay between retries in seconds. Defaults to 1.
            raise_error (bool): raises WaiterConditionWasNotMet.

        Returns:
            self: updated RequestsWaiter instance.

        Raises:
            WaiterConditionWasNotMet: if the condition is not met within the specified
                number of attempts."""
        wait_for_executor(
            executor=self.executor,
            retries=retries,
            delay=delay,
            raise_error=raise_error,
        )
        return self

    def get_result(self) -> Response:
        """Get the final response containing the expected values.

        Returns:
            Response: final response containing the expected values."""
        return self.executor.get_result()


def wait_for_value_in_request(
    request: PreparedRequest | Response,
    status_code: int = 200,
    comparer: COMPARATORS | None = None,
    expected_value: Any = None,
    checker: CHECKERS = "json_checker",
    session: Session | None = None,
    dict_path: str | None = None,
    search_query: str | None = None,
    retries: int = 60,
    delay: int = 1,
) -> Response:
    """Wait for a specified value in a response.

    Args:
        request (PreparedRequest | Response): The request or response to monitor for
            the expected value.
        status_code (int, optional): The expected HTTP status code. Defaults to 200.
        comparer (COMPARATORS | None, optional): The comparer function or operator used
            for value comparison. Defaults to None.
        expected_value (Any, optional): The value to be compared against the response
            data. Defaults to None.
        checker (CHECKERS, optional): The type of checker to use.
        session (Session | None, optional): The requests session to use for sending
            requests. Defaults to None.
        dict_path (str | None, optional): The dot-separated path to the value in the
            response data. Defaults to None.
        search_query (str | None, optional): A search query to use to find the value in
            the response data. Defaults to None.
        retries (int, optional): The number of retries to perform. Defaults to 60.
        delay (int, optional): The delay between retries in seconds. Defaults to 1.

    Returns:
        Response: the final response containing the expected value.

    Raises:
        WaiterConditionWasNotMet: if the condition is not met within the specified
            number of attempts.

    Example:
        To wait for a JSON response where the "status" field equals 200 and request
            returns list of dict.
        ```
            response = wait_for_value_in_requests(
                requests=request,
                status_code=200,
                comparer="have_len_greater",
                expected_value=0,
                checker="json_checker",
                session=session,
                dict_path="data",
            )
        ```"""
    waiter = RequestsWaiter(request=request, status_code=status_code, session=session)

    if comparer:
        waiter.add_checker(
            comparer=comparer,
            checker=checker,
            expected_value=expected_value,
            dict_path=dict_path,
            search_query=search_query,
        )

    return waiter.run(retries=retries, delay=delay).get_result()


def wait_for_values_in_request(
    request: PreparedRequest | Response,
    checkers: list[dict[str, Any]],
    status_code: int = 200,
    session: Session | None = None,
    retries: int = 60,
    delay: int = 1,
) -> Response:
    """Wait for multiple specified values in a response using different checkers.

    Args:
        request (PreparedRequest | Response): The request or response to monitor for the
            expected values.
        checkers (list[dict[str, Any]]): A list of dictionaries, where each dictionary
            contains information about a checker to apply.
            Each dictionary must have keys:
               - checker (CHECKERS): type of checker to use.
               - comparer (COMPARATORS): comparer function or operator used for value
                    comparison.
               - expected_value (Any): the value to be compared against the response
                    data.
               - dict_path (str | None, optional): The dot-separated path to the value
                    in the response data. Defaults to None.
               - search_query (str | None, optional): A search query to use to find the
                    value in the response data. Defaults to None.
        status_code (int, optional): The expected HTTP status code. Defaults to 200.
        session (Session | None, optional): The requests session to use for sending
               requests. Defaults to None.
        retries (int, optional): The number of retries to perform. Defaults to 60.
        delay (int, optional): The delay between retries in seconds. Defaults to 1.

    Returns:
        Response: the final response containing the expected values.

    Raises:
        WaiterConditionWasNotMet: if the condition is not met within the specified
            number of attempts.

    Example:
        To wait for multiple conditions using different checkers:
        ```
           checkers = [
               {
                   "checker": "json_checker",
                   "comparer": "have_len_greater",
                   "expected_value": 0,
                   "dict_path": "data",
               },
               {
                   "checker": "json_checker",
                   "comparer": "is_equal",
                   "expected_value": "success",
                   "search_query": "message",
               }
           ]

           response = wait_for_values_in_request(
               request=request,
               checkers=checkers,
               status_code=200,
               session=session,
               retries=5
           )
        ```"""
    waiter = RequestsWaiter(request=request, status_code=status_code, session=session)

    for checker_dict in checkers:
        waiter.add_checker(**checker_dict)

    return waiter.run(retries=retries, delay=delay).get_result()



from typing import Any, Iterable, Literal, Sized


def is_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is equal to expected_value, False otherwise."""
    return data == expected_value


def is_not_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is not equal to expected_value, False otherwise."""
    return data != expected_value


def is_greater_than(data: Any, expected_value: Any) -> bool:
    """Returns True if data is greater than expected_value, False otherwise."""
    return data > expected_value


def is_lesser_than(data: Any, expected_value: Any) -> bool:
    """Returns True if data is lesser than expected_value, False otherwise."""
    return data < expected_value


def is_greater_than_or_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is greater than or equal to expected_value,
    False otherwise."""
    return data >= expected_value


def is_lesser_than_or_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is lesser than or equal to expected_value,
    False otherwise."""
    return data <= expected_value


def contain(data: Iterable[Any], expected_value: Any) -> bool:
    """Returns True if expected_value is present in data, False otherwise."""
    return expected_value in data


def not_contain(data: Iterable[Any], expected_value: Any) -> bool:
    """Returns True if expected_value is not present in data, False otherwise."""
    return expected_value not in data


def contain_all(data: Iterable[Any], expected_value: Iterable[Any]) -> bool:
    """Returns True if all elements in expected_value are present in data,
    False otherwise."""
    return all((i in data for i in expected_value))


def contain_any(data: Iterable[Any], expected_value: Iterable[Any]) -> bool:
    """Returns True if any element in expected_value is present in data,
    False otherwise."""
    return any((i in data for i in expected_value))


def have_len_equal(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is equal to expected_value,
    False otherwise."""
    return len(data) == expected_value


def have_len_greater(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is greater than expected_value,
    False otherwise."""
    return len(data) > expected_value


def have_len_lesser(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is lesser than expected_value,
    False otherwise."""
    return len(data) < expected_value


COMP_DICT = {
    "is_equal": is_equal,
    "is_not_equal": is_not_equal,
    "is_greater_than": is_greater_than,
    "is_lesser_than": is_lesser_than,
    "is_greater_than_or_equal": is_greater_than_or_equal,
    "is_lesser_than_or_equal": is_lesser_than_or_equal,
    "contain": contain,
    "not_contain": not_contain,
    "contain_all": contain_all,
    "contain_any": contain_any,
    "have_len_equal": have_len_equal,
    "have_len_greater": have_len_greater,
    "have_len_lesser": have_len_lesser,
}

COMPARATORS = Literal[
    "is_equal",
    "is_not_equal",
    "is_greater_than",
    "is_lesser_than",
    "is_greater_than_or_equal",
    "is_lesser_than_or_equal",
    "contain",
    "not_contain",
    "contain_all",
    "contain_any",
    "have_len_equal",
    "have_len_greater",
    "have_len_lesser",
]



from typing import Literal

from .response_checkers import HeadersChecker, JsonChecker

CHECKERS = Literal["json_checker", "headers_checker"]

RESPONSE_CHECKERS = {"json_checker": JsonChecker, "headers_checker": HeadersChecker}



from abc import ABC, abstractmethod
from typing import Any


class Executor(ABC):
    """An abstract base class for defining an executor that can be waited for."""

    @abstractmethod
    def is_condition_met(self) -> bool:
        """Check whether the condition has been met.

        Returns:
            bool: True if the condition has been met, False otherwise."""

    @abstractmethod
    def get_result(self) -> Any:
        """Returns the result of performed actions."""

    @abstractmethod
    def error_message(self) -> str:
        """Return a detailed error message if the condition has not been met."""



import sys

from setuptools import setup

REQUIRED_PYTHON = (3, 10)

if sys.version_info[:2] < REQUIRED_PYTHON:
    sys.stderr.write(
        """\nUnsupported Python version\n.
        This version of bepatient requires at least Python {}.{}""".format(
            *REQUIRED_PYTHON
        )
    )
    sys.exit(1)

setup()



import logging

from requests import PreparedRequest, Response, Session
from requests.exceptions import RequestException

from bepatient.curler import Curler
from bepatient.waiter_src.checker import Checker
from bepatient.waiter_src.checkers.response_checkers import StatusCodeChecker
from bepatient.waiter_src.comparators import is_equal
from bepatient.waiter_src.exceptions.executor_exceptions import ExecutorIsNotReady
from bepatient.waiter_src.executor import Executor

log = logging.getLogger(__name__)


class RequestsExecutor(Executor):
    """An executor that sends a request and waits for a certain condition to be met.
    Args:
        req_or_res (PreparedRequest): The request to send.
        session (Session): The requests session to use."""

    def __init__(
        self,
        req_or_res: PreparedRequest | Response,
        expected_status_code: int,
        session: Session | None = None,
    ):
        self._status_code_checker = StatusCodeChecker(is_equal, expected_status_code)
        self._checkers: list[Checker] = []
        self._failed_checkers: list[Checker] = []
        self._response: Response | None = None

        if session:
            self.session = session
        else:
            log.info("Creating a new Session object")
            self.session = Session()

        if isinstance(req_or_res, Response):
            log.info("Merging response data into session")
            self.session.headers.update(req_or_res.request.headers)
            self._response = req_or_res
            if len(req_or_res.history) > 0:
                self.request = req_or_res.history[0].request
            else:
                self.request = req_or_res.request
        else:
            self.request = req_or_res

    def add_checker(self, checker: Checker):
        """Adds checker function to the list of checkers."""
        self._checkers.append(checker)
        return self

    def is_condition_met(self) -> bool:
        """Sends the request and check if all checkers pass or timeout occurs.

        Returns:
            bool: True if all checkers pass, False otherwise.

        Raises:
            ExecutorIsNotReady: If the executor is not ready to send the request."""
        try:
            self._response = self.session.send(self.request)
        except RequestException:
            log.exception("RequestException! CURL: %s", Curler().to_curl(self.request))
            return False

        if self._status_code_checker.check(self._response):
            self._failed_checkers = [
                checker
                for checker in self._checkers
                if not checker.check(self._response)
            ]
        else:
            self._failed_checkers = [self._status_code_checker]

        if len(self._failed_checkers) == 0:
            return True
        return False

    def get_result(self) -> Response:
        """Returns the response received from the server."""
        if self._response is not None:
            return self._response
        raise ExecutorIsNotReady()

    def error_message(self) -> str:
        """Return a detailed error message if the condition has not been met."""
        if self._response is not None and len(self._failed_checkers) > 0:
            checkers = ", ".join([str(checker) for checker in self._failed_checkers])
            return (
                "The condition has not been met!"
                f" | Failed checkers: ({checkers})"
                f" | {Curler().to_curl(self.request)}"
            )
        raise ExecutorIsNotReady()



from requests import PreparedRequest, Response

from bepatient.curler import Curler


class TestRequestCurl:
    def test_request_to_curl(self, prepared_request: PreparedRequest):
        curl = (
            "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' "
            "https://webludus.pl/"
        )
        assert Curler().to_curl(prepared_request) == curl

    def test_request_without_headers_and_body(self):
        curl = "curl -X GET https://webludus.pl/"
        prepared_request = PreparedRequest()
        prepared_request.prepare(method="get", url="https://webludus.pl")

        assert Curler().to_curl(prepared_request) == curl

    def test_with_utm(self, prepared_request: PreparedRequest):
        prepared_request.url += (  # type: ignore
            "?utm_source=pytest&utm_medium=tests&utm_campaign=bepatient"
        )
        curl = (
            "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' https://"
            "webludus.pl/?utm_source=pytest&utm_medium=tests&utm_campaign=bepatient"
        )

        assert Curler().to_curl(prepared_request) == curl

    def test_post_json(self):
        request = PreparedRequest()
        request.prepare(
            method="post",
            url="https://webludus.pl",
            headers={"task": "test"},
            cookies={"user-token": "abc-123"},
            json={"login": "admin", "password": "admin1"},
        )
        expected_curl = (
            "curl -X POST -H 'task: test' -H 'Cookie: user-token=abc-123'"
            " -H 'Content-Length: 40' -H 'Content-Type: application/json' -d "
            '\'{"login": "admin", "password": "admin1"}\' https://webludus.pl/'
        )

        assert Curler().to_curl(request) == expected_curl

    def test_data(self):
        request = PreparedRequest()
        request.prepare(
            method="post",
            url="https://webludus.pl",
            headers={"task": "test"},
            cookies={"user-token": "abc-123"},
            data={"login": "admin", "password": "admin1"},
        )
        expected_curl = (
            "curl -X POST -H 'task: test' -H 'Cookie: user-token=abc-123'"
            " -H 'Content-Length: 27' -H"
            " 'Content-Type: application/x-www-form-urlencoded'"
            " -d 'login=admin&password=admin1' https://webludus.pl/"
        )

        assert Curler().to_curl(request) == expected_curl

    def test_post_with_charset(self):
        request = PreparedRequest()
        request.prepare(
            method="post",
            url="https://webludus.pl",
            headers={"Content-Type": "text/plain"},
            json={"PL": "ąćęłńóśźż"},
        )
        expected_curl = (
            "curl -X POST -H 'Content-Type: text/plain' -H 'Content-Length: 64'"
            ' -d \'{"PL": "ąćęłńóśźż"}\' https://webludus.pl/'
        )
        assert Curler().to_curl(request, charset="unicode-escape") == expected_curl

    def test_response(self, prepared_request: PreparedRequest):
        prepared_request.headers.update({"content-length": "44"})
        response = Response()
        response.request = prepared_request
        curl = (
            "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123'"
            " https://webludus.pl/"
        )

        assert Curler().to_curl(response) == curl

    def test_response_with_utm(self, prepared_request: PreparedRequest):
        prepared_request.url += (  # type: ignore
            "?utm_source=pytest&utm_medium=tests&utm_campaign=bepatient"
        )
        response = Response()
        response.request = prepared_request
        curl = (
            "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' https://"
            "webludus.pl/?utm_source=pytest&utm_medium=tests&utm_campaign=bepatient"
        )

        assert Curler().to_curl(response) == curl



from shlex import quote
from typing import Any

from requests import PreparedRequest, Response
from requests.structures import CaseInsensitiveDict


class Curler:
    """
    A class to convert a `PreparedRequest` or a `Response` object to a `curl` command.
    """

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _prepare_headers(data: dict[str, Any] | CaseInsensitiveDict[str]) -> str:
        headers = " ".join(
            (f"-H {quote(f'{name}: {value}')}" for name, value in data.items())
        )
        return f" {headers}" if len(headers) > 0 else ""

    @staticmethod
    def _prepare_body(data: str | bytes, charset: str) -> str:
        if isinstance(data, bytes):
            data = data.decode(charset)

        return f" -d {quote(data)}"

    def to_curl(
        self, request: PreparedRequest | Response, charset: str | None = None
    ) -> str:
        """A method to convert a `PreparedRequest` or a `Response` object to a `curl`
        command. Returns the `curl` command as a string.

        Args:
            request (PreparedRequest | Response): The `PreparedRequest` or `Response`
                object to be converted.
            charset (str, optional): The character set to use for encoding the
                request body, if it is a byte string. Defaults to "utf-8".

        Returns:
            the `curl` command as a string."""
        if not charset:
            charset = "utf-8"

        if isinstance(request, Response):
            request = request.request
            if "content-length" in request.headers:
                del request.headers["content-length"]

        curl_command = f"curl -X {quote(request.method)}"  # type: ignore
        curl_command += self._prepare_headers(request.headers)

        if request.body:
            curl_command += self._prepare_body(request.body, charset)

        curl_command += f" {request.url}"
        return curl_command



import logging
from time import sleep

from bepatient.waiter_src.exceptions.waiter_exceptions import WaiterConditionWasNotMet
from bepatient.waiter_src.executor import Executor

log = logging.getLogger(__name__)


def wait_for_executor(
    executor: Executor, retries: int, delay: int, raise_error: bool = True
) -> None:
    """Wait for the given executor to meet its condition.

    Args:
        executor (Executor): The executor to wait for.
        retries (int): The number of times to retry the operation.
        delay (int): The delay in seconds between retries.
        raise_error (bool): raises WaiterConditionWasNotMet

    Raises:
        WaiterConditionWasNotMet: if the condition is not met within the specified
            number of attempts."""
    for attempt in range(retries):
        log.debug(
            "Checking whether the condition has been met. The %s approach", attempt + 1
        )
        if executor.is_condition_met():
            log.debug("Condition met!")
            return
        log.debug("The condition has not been met. Waiting: %s", delay)
        sleep(delay)
    if raise_error:
        raise WaiterConditionWasNotMet(executor.error_message())
